/*
 * Raytracing v2.0
 *
 * Compile and build:
 * zcc +cpm -subtype=pcw80 -compiler=sccz80 -O2 -lm -create-app raytr.c pbmout.c kernel.c screen.c buffered_io.c -o raytr.com
 */
#include <stdio.h>
#include <conio.h>
#include <ctype.h>
#include <graphics.h>
#include <math.h>
#include "kernel.h"
#include "mytypes.h"
#include "pbmout.h"

#define CH_BEEP 7
#define XM 4
#define YM 4
#define SPHERE_COUNT 4
#define XRES 720
#define YRES 256
#define XRES_HALF 360
#define YRES_HALF 128
#define XMAX 719
#define YMAX 255

unsigned char floorColour[] = {0, 3, 6, 11}; // colours to plot the floor
float cameraCoord[] = { 0.3f, -0.5f, 0.0f };
float sphereCoord[SPHERE_COUNT][3] = {
        { -0.8, -1.0,  3.2 },
        {  0.0, -0.45, 2.0 },
        {  1.2, -0.7,  2.5 },
        {  0.4, -1.0,  4.0 }
    };
float radius[] = { 0.7, 0.3, 0.5, 0.4 };
float q[4];
float x, y, z;
float dx, dy, dz;

// function prototypes
void trace(int i, int j);
float xMinusIntX(float f);
void plotGrey(int x, int y, unsigned char r);
void beep(void);
void invertInMemory(void);

/*
 * main
 */
int main(int argc, char *argv[])
{
    int i, j;
	char ch;
	char *filename;
    
	gx_init();
    clg();
    printf("%cH", 27); // ESC H = VT100 cursor home
    printf("%c0", 27); // ESC 0 = switch off status bar

	// filename to save the picture to
	if (argc > 1) {
		filename = argv[1];
	}
	else {
		filename = "scene.pbm";
	}

    for (i = 0; i < 4; ++i) {
        q[i] = radius[i] * radius[i];
    }
    
    for (i = 0; i < YRES * 2; i += YM) {
        for (j = 0; j < XRES; j += XM) {
            x = cameraCoord[0];
            y = cameraCoord[1];
            z = cameraCoord[2];
            dx = (float) (j - XRES_HALF);
            dy = (float) (YRES - i);
            dz = 450.0;
            trace(i, j);
        }
    }

    beep();
	
	do {
		ch = getch();
		ch = toupper(ch);
		switch (ch) {
			case 'I': beep();
			          invertInMemory();
					  beep();
			          break;
			case 'W': beep();
			          saveScreenToPBM(filename, true);
					  beep();
			          break;
		}
	} while (ch != 'Q' && ch != 'X');
	
	beep();
    clg();
    printf("%c1", 27); // ESC 1 = switch on status bar
    printf("%ce", 27); // ESC e = show cursor
    
    return 0;
}

/*
 * Traces a primary ray.
 */
void trace(int i, int j)
{
    float dd;
    float s;
    float px, py, pz;
    float pp;
    float sc;
    float aa, bb;
    char n;
    unsigned char k, kn;
    unsigned char r1;
    unsigned char ba;
    float r1f;
    float nx, ny, nz;
    float nn;
    float l;
    float u, v;
    
    dd = dx * dx + dy * dy + dz * dz;
    s = 0.0;
    ba = 2;
    
    for (;;) {
        if (y >= 0 || dy <= 0) {
            n = -1;
        }
        else {
            n = 0;
        }
		
        if (!n) {
            s = -y / dy;
        }
		
        for (k = 0; k < SPHERE_COUNT; ++k) {
            px = sphereCoord[k][0] - x;
            py = sphereCoord[k][1] - y;
            pz = sphereCoord[k][2] - z;
            pp = px * px + py * py + pz * pz;
            sc = px * dx + py * dy + pz * dz;
			
            if (sc > 0) {
                bb = sc * sc / dd;
                aa = q[k] - pp + bb;
                if (aa > 0) {
                    sc = (sqrt(bb) - sqrt(aa)) / sqrt(dd);
                    if (sc < s || n < 0) {
                        kn = k;
						n = 1;
                        s = sc;
                    }
                }
            }
        }
		
        if (n < 0) {
            r1f = dy * dy / dd * 11.0; // 26.0;
            r1 = (unsigned char) r1f;
            ++r1;
            plotGrey(j, i, r1);
            n = 0;
            return;
        }
		
        dx *= s;
        dy *= s;
        dz *= s;
        dd *= s * s;
        x += dx;
        y += dy;
        z += dz;
		
        if (n == 0) {
            break;
        }
		
        nx = x - sphereCoord[kn][0];
        ny = y - sphereCoord[kn][1];
        nz = z - sphereCoord[kn][2];
        nn = nx * nx + ny * ny + nz * nz;
        l = 2 * (dx * nx + dy * ny + dz * nz) / nn;
        dx -= nx * l;
        dy -= ny * l;
        dz -= nz * l;
    }

    for (k = 0; k < SPHERE_COUNT; ++k) {
        u = sphereCoord[k][0] - x;
        v = sphereCoord[k][2] - z;
        if (u * u + v * v <= q[k]) {
            ba = 0;
        }
    }
    
    if ((xMinusIntX(x) > 0.5) == (xMinusIntX(z) > 0.5)) {
        r1 = floorColour[ba];
    }
    else {
        r1 = floorColour[ba + 1];
    }
	
    plotGrey(j, i, r1);
}

float xMinusIntX(float f)
{
    int i;

	if (f < 0) {
		f = f + 0.000001; // avoid rounding bug
		f = -f;
		i = (int) f;
		return 1.0 - f + (float) i;
	}
	else {
		i = (int) f;
		return f - (float) i;
	}
}

void plotGrey(int x, int y, unsigned char r)
{
    y = YMAX - (y >> 1);
    switch(r) {
        case  0:
            break;
        case  1:
            plot(++x, y);
            break;
        case  2:
            plot(++x, y);
            plot(++x, ++y);
            break;
        case  3:
            plot(x++, y);
            plot(++x, y);
            plot(--x, ++y);
            break;
        case  4:
            plot(x++, y);
            plot(++x, y);
            plot(--x, ++y);
            ++x;
            plot(++x, y);
            break;
        case  5:
            plot(x++, y);
            ++x;
            plot(++x, y);
            --x;
            plot(--x, ++y);
            plot(++x, y);
            break;
        case  6:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(--x, ++y);
            ++x;
            plot(++x, y);
            break;
        case  7:
            plot(x, y);
            plot(++x, y);
            ++x;
            plot(++x, y);
            --x;
            plot(--x, ++y);
            plot(++x, y);
            break;
        case  8:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(++x, y);
            --x;
            plot(--x, ++y);
            plot(++x, y);
            break;
        case  9:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(--x, ++y);
            plot(++x, y);
            plot(++x, y);
            break;
        case 10:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(++x, y);
            plot(--x, ++y);
            plot(--x, y);
            plot(--x, y);
            break;
        case 11:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(++x, y);
            plot(x, ++y);
            plot(--x, y);
            plot(--x, y);
            plot(--x, y);
            break;
        default:
            printf("%d ", r);
    }
}

// Emits a beep through the loudspeaker.
void beep()
{
    putchar(CH_BEEP);
}
